import requests
import aurorax
import pprint
import json
from typing import Optional, Dict, Any, List, Union
from pydantic import BaseModel
from .._internal.util import json_converter

# reqest globals
DEFAULT_RETRIES = 2
REQUEST_HEADERS = {
    "accept": "application/json",
    "Content-Type": "application/json"
}
API_KEY_HEADER_NAME = "x-aurorax-api-key"

# private dynamic globals
__api_key = ""


def get_api_key():
    return __api_key


def authenticate(api_key: str) -> int:
    """
    Set authentication values for use with subsequent queries

    :param api_key: api key
    :type api_key: str

    :return: 0
    :rtype: int
    """

    # set the global variable
    global __api_key
    __api_key = api_key

    # return
    return 0


class AuroraXResponse(BaseModel):
    request: Any
    data: Any
    status_code: int

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return pprint.pformat(self.__dict__)


class AuroraXRequest(BaseModel):
    url: str
    method: str
    params: Optional[Dict] = {}
    body: Union[Optional[Dict], Optional[List]] = {}
    headers: Optional[Dict] = {}
    null_response: Optional[bool] = False

    def __merge_headers(self):
        # set initial headers
        all_headers = {
            "accept": "application/json",
            "Content-Type": "application/json"
        }

        # add headers passed into the class
        for key, value in self.headers.items():
            all_headers[key] = value

        # add api key
        api_key = get_api_key()
        if (api_key is not None):
            all_headers[API_KEY_HEADER_NAME] = api_key

        # return
        return all_headers

    def execute(self, limited_evaluation: bool = False) -> AuroraXResponse:
        """Execute an AuroraX request

        :param limited_evaluation: set true if you don't want to evaluate the response outside of
                                   the retry mechanism, defaults to False
        :type limited_evaluation: bool, optional

        :raises aurorax.AuroraXMaxRetriesException: max retry error
        :raises aurorax.AuroraXUnexpectedContentTypeException: unexpected content error

        :return: response
        :rtype: AuroraXResponse
        """
        # sanitize data
        body_santized = json.dumps(self.body, default=json_converter)

        # make request
        req = requests.request(self.method,
                               self.url,
                               headers=self.__merge_headers(),
                               params=self.params,
                               data=body_santized)

        # retry request if needed
        for i in range(0, aurorax.api.DEFAULT_RETRIES):
            if (req.status_code == 500 and "text/plain" in req.headers["Content-Type"]):
                if (i == (aurorax.api.DEFAULT_RETRIES - 1)):
                    raise aurorax.AuroraXMaxRetriesException("%s (%s)" % (req.content.decode(),
                                                                          req.status_code))
                req = requests.request(self.method,
                                       self.url,
                                       headers=self.__merge_headers(),
                                       params=self.params,
                                       json=self.body,
                                       data=body_santized)
            else:
                break

        # check if authorization worked
        if (req.status_code == 401):
            raise aurorax.AuroraXUnauthorizedException("%s %s" % (req.status_code,
                                                                  req.json()["error_message"]))

        # check if we only want to do limited evaluation
        if (limited_evaluation is True):
            res = AuroraXResponse(request=req, data=None, status_code=req.status_code)
            return res

        # check content type
        if (self.null_response is False):
            if (req.headers["Content-Type"] == "application/json"):
                response_data = req.json()
            else:
                raise aurorax.AuroraXUnexpectedContentTypeException("%s (%s)" % (req.content.decode(),
                                                                                 req.status_code))
        else:
            if (req.status_code != 200 and req.status_code != 201 and req.status_code != 202):
                response_data = req.json()
            else:
                response_data = None

        # check for server error
        if (req.status_code == 500):
            raise aurorax.AuroraXException("%s (%s)" % (req.json()["error_message"],
                                                        req.status_code))

        # create reponse object
        res = AuroraXResponse(request=req, data=response_data, status_code=req.status_code)

        # return
        return res

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return pprint.pformat(self.__dict__)



import datetime
import pprint
import aurorax
from pydantic import BaseModel
from typing import Dict, List


class DataProduct(BaseModel):
    """
    DataProduct data type

    :param identifier: data source ID
    :type identifier: int
    :param program: program name
    :type program: str
    :param platform: platform name
    :type platform: str
    :param instrument_type: instrument type name
    :type instrument_type: str
    :param data_product_type: data product type (keogram, movie, summary_plot)
    :type instrument_type: str
    :param start: starting timestamp for the record in UTC
    :type start: datetime.datetime
    :param end: ending timestamp for the record in UTC
    :type end: datetime.datetime
    :param url: URL location of data product
    :type url: str
    :param metdata: metadata values for this record
    :type metdata: Dict
    """
    identifier: int
    program: str
    platform: str
    instrument_type: str
    data_product_type: str
    start: datetime.datetime
    end: datetime.datetime
    url: str
    metadata: Dict

    def to_json_serializable(self) -> Dict:
        """
        Convert object to a JSON-serializable object (ie. translate datetime
        objects to strings)

        :return: dictionary JSON-serializable object
        :rtype: Dict
        """
        d = self.__dict__

        # format epoch as str
        if (type(d["start"]) is datetime.datetime):
            d["start"] = d["start"].strftime("%Y-%m-%dT%H:%M:00.000Z")
        if (type(d["end"]) is datetime.datetime):
            d["end"] = d["end"].strftime("%Y-%m-%dT%H:%M:00.000Z")

        # format metadata
        if (type(self.metadata) is dict):
            for key, value in self.metadata.items():
                if (type(value) is datetime.datetime or type(value) is datetime.date):
                    self.metadata[key] = self.metadata[key].strftime("%Y-%m-%dT%H:%M:%S.%f")
        if (type(self.metadata) is list):
            self.metadata = {}

        # return
        return d

    def __str__(self) -> str:
        """
        String method

        :return: string format
        :rtype: str
        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation

        :return: object representation
        :rtype: str
        """
        return pprint.pformat(self.__dict__)


def upload(identifier: int, records: List["DataProduct"]) -> int:
    """
    Upload data product records to AuroraX

    :param identifier: data source ID
    :type identifier: int
    :param records: data product records to upload
    :type records: List[DataProduct]

    :raises aurorax.AuroraXMaxRetriesException: max retry error
    :raises aurorax.AuroraXUnexpectedContentTypeException: unexpected content error
    :raises aurorax.AuroraXUploadException: upload error

    :return: 0 for success, raises exception on error
    :rtype: int
    """
    # translate each data product record to a request-friendly dict (ie. convert datetimes to strings, etc.)
    for i, _ in enumerate(records):
        if (type(records[i]) is DataProduct):
            records[i] = records[i].to_json_serializable()

    # make request
    url = aurorax.api.urls.data_products_upload_url.format(identifier)
    req = aurorax.AuroraXRequest(method="post", url=url, body=records, null_response=True)
    res = req.execute()

    # evaluate response
    if (res.status_code == 400):
        raise aurorax.AuroraXUploadException("%s - %s" % (res.data["error_code"], res.data["error_message"]))

    # return
    return 0



import aurorax
from typing import List, Dict, Optional


def list(order: Optional[str] = "identifier") -> List:
    """
    Retrieve all data source records

    :param order: value to order results by (identifier, program, platform,
                  instrument_type, display_name, owner), defaults to "identifier"
    :type format: Optional[str], optional

    :raises aurorax.AuroraXMaxRetriesException: max retry error
    :raises aurorax.AuroraXUnexpectedContentTypeException: unexpected error

    :return: all data sources
    :rtype: List
    """
    # make request
    req = aurorax.AuroraXRequest(method="get", url=aurorax.api.urls.data_sources_url)
    res = req.execute()

    # order results
    res.data = sorted(res.data, key=lambda x: x[order])

    # return
    return res.data


def get(program: str,
        platform: str,
        instrument_type: str,
        format: Optional[str] = "basic_info") -> Dict:
    """
    Retrieve a specific data source record

    :param program: program
    :type program: str
    :param platform: program
    :type platform: str
    :param instrument_type: program
    :type instrument_type: str
    :param format: record format, defaults to "basic_info"
    :type format: Optional[str], optional

    :raises aurorax.AuroraXMaxRetriesException: max retry error
    :raises aurorax.AuroraXUnexpectedContentTypeException: unexpected error

    :return: data source
    :rtype: Dict
    """
    # make request
    params = {
        "program": program,
        "platform": platform,
        "instrument_type": instrument_type,
        "format": format,
    }
    req = aurorax.AuroraXRequest(method="get", url=aurorax.api.urls.data_sources_url, params=params)
    res = req.execute()

    # set results to the first thing
    if (len(res.data) == 1):
        res.data = res.data[0]
    else:
        res.data = {}

    # return
    return res.data


def get_using_filters(program: Optional[str] = None,
                      platform: Optional[str] = None,
                      instrument_type: Optional[str] = None,
                      source_type: Optional[str] = None,
                      owner: Optional[str] = None,
                      format: Optional[str] = "basic_info",
                      order: Optional[str] = "identifier",) -> List:
    """
    Retrieve all data source records matching a filter

    :param program: program, defaults to None
    :type program: Optional[str], optional
    :param platform: program, defaults to None
    :type platform: Optional[str], optional
    :param instrument_type: program, defaults to None
    :type instrument_type: Optional[str], optional
    :param source_type: program, defaults to None
    :type source_type: Optional[str], optional
    :param owner: program, defaults to None
    :type owner: Optional[str], optional
    :param format: record format, defaults to "basic_info"
    :type format: Optional[str], optional
    :param order: value to order results by (identifier, program, platform,
                  instrument_type, display_name, owner), defaults to "identifier"
    :type order: Optional[str], optional

    :raises aurorax.AuroraXMaxRetriesException: max retry error
    :raises aurorax.AuroraXUnexpectedContentTypeException: unexpected error

    :return: matching data sources
    :rtype: List
    """
    # make request
    params = {
        "program": program,
        "platform": platform,
        "instrument_type": instrument_type,
        "source_type": source_type,
        "owner": owner,
        "format": format,
    }
    req = aurorax.AuroraXRequest(method="get", url=aurorax.api.urls.data_sources_url, params=params)
    res = req.execute()

    # order results
    res.data = sorted(res.data, key=lambda x: x[order])

    # return
    return res.data


def get_using_identifier(identifier: int, format: Optional[str] = "basic_info") -> Dict:
    """
    Retrieve all data source records matching a filter

    :param identifier: data source identifier
    :type identifier: int
    :param format: record format, defaults to "basic_info"
    :type format: Optional[str], optional

    :raises aurorax.AuroraXMaxRetriesException: max retry error
    :raises aurorax.AuroraXUnexpectedContentTypeException: unexpected error

    :return: matching data sources
    :rtype: List
    """
    # make request
    params = {
        "format": format,
    }
    url = "%s/%d" % (aurorax.api.urls.data_sources_url, identifier)
    req = aurorax.AuroraXRequest(method="get", url=url, params=params)
    res = req.execute()

    # return
    return res.data


def get_stats(identifier: int,
              format: Optional[str] = "basic_info",
              slow_method: Optional[bool] = False) -> Dict:
    """
    Retrieve statistics for a data source

    :param identifier: data source identifier
    :type identifier: int
    :param format: record format, defaults to "basic_info"
    :type format: Optional[str], optional
    :param slow_method: use slow method which gets most up-to-date stats info, defaults to "False"
    :type slow_method: Optional[bool], optional

    :raises aurorax.AuroraXMaxRetriesException: max retry error
    :raises aurorax.AuroraXUnexpectedContentTypeException: unexpected error

    :return: stats info
    :rtype: Dict
    """
    # make request
    params = {
        "format": format,
        "slow_method": slow_method,
    }
    url = "%s/%d/stats" % (aurorax.api.urls.data_sources_url, identifier)
    req = aurorax.AuroraXRequest(method="get", url=url, params=params)
    res = req.execute()

    # return
    return res.data


def add(program: str,
        platform: str,
        instrument_type: str,
        source_type: str,
        display_name: str,
        ephemeris_metadata_schema: List[Dict] = [],
        data_product_metadata_schema: List[Dict] = [],
        maintainers: List = [],
        identifier: int = None) -> Dict:
    """
    Add new data source to AuroraX

    :param program: program name
    :type program: str
    :param platform: platform name
    :type platform: str
    :param instrument_type: instrument type
    :type instrument_type: str
    :param source_type: source type (heo, leo, lunar, ground)
    :type source_type: str
    :param display_name: a more user-friendly display name
    :type display_name: str
    :param ephemeris_metadata_schema: metadata schema, defaults to []
    :type ephemeris_metadata_schema: List[Dict], optional
    :param data_product_metadata_schema: metadata schema, defaults to []
    :type data_product_metadata_schema: List[Dict], optional
    :param maintainers: list of users to give maintainer permissions to, defaults to []
    :type maintainers: List, optional
    :param identifier: data source ID, defaults to None
    :type identifier: int, optional
    :return: the created data source record details
    :rtype: Dict

    :raises aurorax.AuroraXMaxRetriesException: max retry error
    :raises aurorax.AuroraXUnexpectedContentTypeException: unexpected error
    :raises aurorax.AuroraXDuplicateException: duplicate data source, already exists

    :return: created data source
    :rtype: Dict
    """
    # do request
    request_data = {
        "program": program,
        "platform": platform,
        "instrument_type": instrument_type,
        "source_type": source_type,
        "display_name": display_name,
        "ephemeris_metadata_schema": ephemeris_metadata_schema,
        "data_product_metadata_schema": data_product_metadata_schema,
        "maintainers": maintainers,
    }
    if (identifier is not None):
        request_data["identifier"] = identifier
    req = aurorax.AuroraXRequest(method="post", url=aurorax.api.urls.data_sources_url, body=request_data)
    res = req.execute()

    # evaluate response
    if (res.status_code == 409):
        raise aurorax.AuroraXDuplicateException("%s - %s" % (res.data["error_code"], res.data["error_message"]))

    # return
    return res.data


def delete(identifier: int) -> int:
    """
    Delete a data source

    :param identifier: data source identifier
    :type identifier: int

    :raises aurorax.AuroraXMaxRetriesException: max retry error
    :raises aurorax.AuroraXUnexpectedContentTypeException: unexpected error
    :raises aurorax.AuroraXNotFoundException: data source not found
    :raises aurorax.AuroraXConflictException: conflict of some type

    :return: 0 on success
    :rtype: int
    """
    # do request
    url = "%s/%d" % (aurorax.api.urls.data_sources_url, identifier)
    req = aurorax.AuroraXRequest(method="delete", url=url, null_response=True)
    res = req.execute()

    # evaluate response
    if (res.status_code == 404):
        raise aurorax.AuroraXNotFoundException("%s - %s" % (res.data["error_code"], res.data["error_message"]))
    elif (res.status_code == 409):
        raise aurorax.AuroraXConflictException("%s - %s" % (res.data["error_code"], res.data["error_message"]))

    # return
    return res.data


def update(identifier: int,
           program: Optional[str] = None,
           platform: Optional[str] = None,
           instrument_type: Optional[str] = None,
           source_type: Optional[str] = None,
           display_name: Optional[str] = None,
           ephemeris_metadata_schema: Optional[List[Dict]] = None,
           data_product_metadata_schema: Optional[List[Dict]] = None,
           metadata: Optional[Dict] = None) -> Dict:
    """
    Update a data source in AuroraX

    :param identifier: data source identifier
    :type identifier: int
    :param program: new program name, defaults to None
    :type program: Optional[str], optional
    :param platform: new platform name, defaults to None
    :type platform: Optional[str], optional
    :param instrument_type: new instrument type, defaults to None
    :type instrument_type: Optional[str], optional
    :param source_type: new source type (heo, leo, lunar, ground), defaults to None
    :type source_type: Optional[str], optional
    :param display_name: new user-friendly display name, defaults to None
    :type display_name: Optional[str], optional
    :param ephemeris_metadata_schema: new metadata schema, defaults to []
    :type ephemeris_metadata_schema: List[Dict], optional
    :param data_product_metadata_schema: new metadata schema, defaults to []
    :type data_product_metadata_schema: List[Dict], optional
    :param metadata: new metadata, defaults to None
    :type metadata: Optional[Dict], optional

    :raises aurorax.AuroraXMaxRetriesException: max retry error
    :raises aurorax.AuroraXUnexpectedContentTypeException: unexpected error
    :raises aurorax.AuroraXNotFoundException: data source not found

    :return: updated data source
    :rtype: Dict
    """
    # get the data source first
    ds = get_using_identifier(identifier)
    if (ds == {}):
        raise aurorax.AuroraXNotFoundException("data source not found")

    # set URL
    url = "%s/%d" % (aurorax.api.urls.data_sources_url, identifier)

    # replace data source values with ones passed into this function
    if (program is not None):
        ds["program"] = program
    if (platform is not None):
        ds["platform"] = platform
    if (instrument_type is not None):
        ds["instrument_type"] = instrument_type
    if (source_type is not None):
        ds["source_type"] = source_type
    if (display_name is not None):
        ds["display_name"] = display_name
    if (ephemeris_metadata_schema is not None):
        ds["ephemeris_metadata_schema"] = ephemeris_metadata_schema
    if (data_product_metadata_schema is not None):
        ds["data_product_metadata_schema"] = data_product_metadata_schema
    if (metadata is not None):
        ds["metadata"] = metadata

    # make request
    req = aurorax.AuroraXRequest(method="put", url=url, body=ds)
    res = req.execute()

    # return
    return res.data
